from pyquicksql.utils.sql_errors import *
from pyquicksql.utils.sql_enums import *

import pymysql as sql

def commit(data_required, data_specific, column, element, indentifiers, identifier_elements):
	'''
		(list of strings, list of strings, list of strings, list of strings) -> (boolean)
		@conditions valid data_required and data_specific list of strings must be passed as
					arguments to the fucntion in acordance to the outlines bellow. Furthermore,
					columns and elements must be of length 1 (one) and all doc-string types must
					be followed to ensure the function runs without bugs.
		@returns True if the elements were sucesfully appended to the columns of pre-existing rows,
				 otherwise if one of the elements is pre-existing, the elements have been appeneded
				 to the database without interuption.
		@exception if invalid input was encountered, the function will return False with the assumption
				   there is a failure to acknowledge or follow the outlines by the docstrings above
				
			*** SHARED AMONGST VARIOUS DATABASE CALLS : ONE SERVER (MOST LIKELY) ***
			data_required = [
				0 : IP ADDRESS OF DATABASE
				1 : PORT OF DATABASE
				2 : USERNAME FOR DB ACESS
				3 : PASSWORD FOR DB ACESS
			]
		
			*** NOT SPECIFIC TO VARIOUS CALLS : HIGH LIKELYHOOD FOR VARIATIONS AMONGST CALLS ***
			data_specific = [
				0 : AUTOCOMMIT (BOOL)
				1 : DB NAME
				2 : DB TABLE NAME
			]
			
			unknown_data = [
				0 : column
				1 : element
			]
			
			known_data = [
				0 : column
				1 : element
			]
			
			*** GENERIC SQL QUERY TO MODIFY THE DATA WITHIN THE LOCAL SERVERS DATABASE ***
	'''
	connection = sql.connect(
		host = data_required.getIP(),
		port = data_required.getPort(),
		user = data_required.getAccount()[0],
		passwd = data_required.getAccount()[1],
		autocommit = (data_specific.getAutocommit() == 'True'),
		db = data_specific.getName(),
	)

	try:
		with connection.cursor() as cursor:
			#the above code defines the type of request made to the sql server, depending
			#on whether the ipaddress has an associated id or not, hence we execute it now
			identifiers = ','.join(indentifiers)
			identifier_elements = '\',\''.join(identifier_elements)
			s = f"UPDATE {data_specific[2]} SET {column} = \'{element}\' WHERE ({indentifiers}) = (\'{identifier_elements}\');"
			cursor.execute(s)
			connection.commit()
			#while no changes are being made, make sure that the connection to
			#the sql server is closed and not left hanging (unforseen errors)
			connection.close()
	except:
		raise ElementNotFound()
		return False
	finally:
		#in the event the code is run safley and without bugs or errors; it is safe to
		#assume and return to the main logical flow that the appropriate requested data
		#has been added or modified on the servers sql database
		return True


from pyquicksql.utils.sql_errors import *
from pyquicksql.utils.sql_enums import *

from pyquicksql.mysql import mysql_lookup

import pymysql as sql

def commit(data_required, data_specific, columns, elements):
	'''
		(list of strings, list of strings, list of strings, list of strings) -> (boolean)
		@conditions valid data_required and data_specific list of strings must be passed as
					arguments to the fucntion in acordance to the outlines bellow. Furthermore,
					columns and elements must be of length 1 (one) and all doc-string types must
					be followed to ensure the function runs without bugs.
		@returns True if the elements were sucesfully appended to the columns of pre-existing rows,
				 otherwise if one of the elements is pre-existing, the elements have been appeneded
				 to the database without interuption.
		@exception if invalid input was encountered, the function will return False with the assumption
				   there is a failure to acknowledge or follow the outlines by the docstrings above
				
			*** SHARED AMONGST VARIOUS DATABASE CALLS : ONE SERVER (MOST LIKELY) ***
			data_required = [
				0 : IP ADDRESS OF DATABASE
				1 : PORT OF DATABASE
				2 : USERNAME FOR DB ACESS
				3 : PASSWORD FOR DB ACESS
			]
		
			*** NOT SPECIFIC TO VARIOUS CALLS : HIGH LIKELYHOOD FOR VARIATIONS AMONGST CALLS ***
			data_specific = [
				0 : AUTOCOMMIT (BOOL)
				1 : DB NAME
				2 : DB TABLE NAME
			]
	'''
	connection = sql.connect(
		host = data_required.getIP(),
		port = data_required.getPort(),
		user = data_required.getAccount()[0],
		passwd = data_required.getAccount()[1],
		autocommit = (data_specific.getAutocommit() == 'True'),
		db = data_specific.getName(),
	)
	
	#check to make sure the same number of elements and columns are given
	if (len(columns) != len(elements)):
		return False
	
	#check to see if the information exists
	check = []
	for i in range(0, len(columns)):
		check.append(mysql_lookup.retreive(data_required, data_specific, columns[i], columns[i], elements[i]))
	
	try:
		check.index('') #if not, input the values like a new entry
		#in the table 'user_lookup' enter a new row of data acording to the respective
		#colums of id (user-id) and ipaddr(the ip-address that is requesting to be
		#associated globaly with that user-id amongst all users in the networks)
		table = data_specific.getTable()
		columns = ','.join(columns)
		elements = '\',\''.join(elements)
		s = f'INSERT INTO {table}({columns}) VALUES (\'{elements}\');'
	except: #else, make modifications to already existing data for the associated ip
		#IF THE ID ALREADY EXISTS (We would arrise problems if we have repeated data
		#in the sql table so we need to AVOID THIS AT ALL COSTS) aka. where we find
		#the same ip-address' row in the database update the old user-id with the new
		#one requested by the user
		return False

	try:
		with connection.cursor() as cursor:
			#the above code defines the type of request made to the sql server, depending
			#on whether the ipaddress has an associated id or not, hence we execute it now
			cursor.execute(s)
			connection.commit()
			#while no changes are being made, make sure that the connection to
			#the sql server is closed and not left hanging (unforseen errors)
			connection.close()
	except:
		raise ElementAlreadyExists()
		return False
	finally:
		#in the event the code is run safley and without bugs or errors; it is safe to
		#assume and return to the main logical flow that the appropriate requested data
		#has been added or modified on the servers sql database
		return True


class Server:
	def __init__(self, _name, _ip, _port, _username, _password):
		'''
			(Name, string, int) -> None
			initializes a new database class to be inputed into a connect class
			@paramaters a Name enum (MYSQL, SQLITE) must be at least specified
			@exception the default type will be set to MYSQL as it is the more popular choice
		'''
		self._name = _name
		self._ip = _ip
		self._port = _port
		self._username = _username
		self._password = _password
	def getName(self):
		'''
			(Server) -> (Name)
			the getter function for the enum type of the server
		'''
		return self._name
	def getIP(self):
		'''
			(Server) -> (string)
			the getter function for the ip string of the connection
		'''
		return self._ip
	def getPort(self):
		'''
			(Server) -> (int)
			the getter function for the port of the connection
		'''
		return self._port
	def getAccount(self):
		'''
			(Server) -> (tuple)
			the getter function for the account (username, password) of the server
		'''
		return ( self._username, self._password )
	def setName(self, new_name):
		'''
			(Server, Name) -> None
			@params new_name must be of enum type Name
			@exception if the params arn't follow no change will be made
		'''
		if ( self._name == Name.MYSQL or self._name == Name.SQLITE ):
			self._name = new_name
	def setIP(self, new_ip):
		'''
			(Server, string) -> None
			@params new_ip must be of type string to rep. the ip
			@exception the ip will not be changed
		'''
		if ( isinstance(new_ip, str) ):
			self._ip = new_ip
	def setPort(self, new_port):
		'''
			(Server, int) -> None
			@params new_port must be of type int to rep. the port
			@exception the port will not be changed
		'''
		if ( isinstance(new_port, int) ):
			self._port = new_port
	def __eq__(self, other):
		'''
			(Server) -> (boolean)
			compares the current class to other on a basis of the Name enum type, ip, and port
		'''
		if ( self._name != other._name ):
			return False
		if ( self._ip != other._ip ):
			return False
		if ( self._port != other._port ):
			return False
	def __repr__(self):
		'''
			(Server) -> (string)
			Returns a string representation of the Database class structure
		'''
		return f'Server({self.name}, {self.ip}, {self.port})'
	def __str__(self):
		'''
			(Server) -> (string)
			Returns the a visualy apealing string representation of the Database.
		'''
		return f'{self.name} -> {self.ip}:{self.port}'


##########################################
#		Generic Logger Errors
##########################################

class DirectoryNotFound(Exception):
	def __init__(self, directory):
		Exception.__init__(self, f"Logger Error: the file in _directory is error-prone\nProvided with: {directory}")
		self.directory = directory

class LogNotFound(Exception):
	def __init__(self, directory):
		Exception.__init__(self, f"Logger Error: the log element requested in _directory does not exist\nProvided with: {directory}")
		self.directory = directory
		
##########################################
#		Logger Markup Errors
##########################################

class MarkupSyntaxError(Exception):
	def __init__(self, text):
		Exception.__init__(self, f"Markup Language Error: the provided text does not follow the logger syntax : {text}")


from enum import Enum

class Name(Enum):
	MYSQL = 1
	SQLITE = 2

class Status(Enum):
	CONNECTED = 1 #the message is being sent
	DISCONECTED = 2 #the message was dropped by the server
	DELAYED = 3 #the message has gone on for a suspiciously long time
	DROPPED = 4 #the connection was dropped by the program

class Message(Enum):
	UNSECURE = 1 #No encryption, no SQL injection check
	ENCRYPTED = 2 #encrypted, No SQL injection check
	SECURE = 3 # not encrypted, SQL injection check
	VERY_SECURE = 4 #encrypted, SQL injection check


##########################################
#	Known Encounterable Errors
##########################################

class ElementNotFound(Exception):
	def __init__(self, SQLElements):
		Exception.__init__(self, f"Database Error: could not find an element with the given paramaters: {SQLElements}")
		self.SQLElements = SQLElements
	
class ElementAlreadyExists(Exception):
	def __init__(self, SQLElements):
		Exception.__init__(self, "Database Error: while attempting to push to the database, one was found with the params: {SQLElements}")
		self.SQLElements = SQLElements

##########################################
#	Ambigous SQL Errors
##########################################	

class ConnectionError(Exception):
	def __init__(self, SQLElements):
		Exception.__init__(self, "Database Error: there was an error while atempting to modify a set of data")
		self.SQLElements = SQLElements


from cffi import FFI
from datetime import datetime

from pyquicksql.utils.logger_enums import Column
from pyquicksql.utils.logger_errors import *

'''
	the following code is a string parser for a bitstream received by a python socket
	@returns the server, database type, and further messages based on the respective req
	0, 1, 2, 3 as located in the bitstream b'[msg][req type][status][location/userid]'
	@example the bitsream b'<toString(server)<>toString(database)>!message!' is received
	by the function with the request type '0', the returned string will be parsed into the
	string representation of the server passed into the logger.
'''

class Logger:
	
	def __init__(self, _directory='default/general_logs.txt'):
		'''
			(Logger, string) -> None
			the initializer function for the Logger class
			
			@paramaters the _directory be a string to a propper folder
			@exception if no _directory string is given it will go to the
					   default logger location 'general/general_logs.txt'
		'''
		#create a new file with the given directory
		createDirectory = open(_directory, 'w+')
		createDirectory.close()
		
		self._directory = _directory
		self._logs = []
		self._commits = 0
	
	def parse(self, line, column):
		'''
			(string, Column) -> (string)
			parses the formated text: 
				'<toString(server)<>toString(database)>!message!'
			
			@paramaters the request (int argument) must be a valid
						integer between 0 and 2 representing the
						various pieces of data present within the
						bitsream:
							a) int 0 : the type of database server
							b) int 1 : the database, table, autocommit
							c) int 2 : the specific message of the log
			@returns the datatype associated with the integer provided
					 by the function argument
			@exception returns None type if an unsupported bitsream is
					   provided by the user
		'''
		ffi = FFI()
		lib = ffi.dlopen("data_markup/target/release/liblibdatamarkup.dylib")

		ffi.cdef('char* parse(const char *n, int);')
		
		try:
			request_val = ffi.new('char[]', bytes(line.decode(), 'utf-8'))
			stream_modified = ffi.string(lib.parse(request_val, column)).decode('utf-8')
			return stream_modified
		except Exception as e:
			raise MarkupSyntaxError(line)
	
	def getDirectory(self):
		'''
			(Logger) -> (string)
			the getter function for the directory of all the log files
			
			@returns the full directory in the format of a string
		'''
		return self._directory
	
	def getCommits(self):
		'''
			(Logger) -> (int)
			the getter function for the number of commited logs during the scripts runtime
			
			@returns an integer values >= 0 representing the number of commits during the session
		'''
		return self._commits
	
	def getLogs(self):
		'''
			(Logger) -> (list of strings)
			the getter function for the temporary variable for logs within the current session
			
			@returns a list of strings of length n representing the number of logs commited to the
					 folder during the current session (runtime of the script)
		'''
		return self._logs
	
	def copy(self):
		'''
			(Directory) -> (boolean)
			coppies the class file _directory logs to the class list var _logs
			
			@paramaters the _directory string is a real file with the correct path
			@exception the function will throw a DirectoryNotFoundError
		'''
		self._logs = []
		try:
			file_logs = open(self._directory, 'r')
			line = file_logs.readLine()
			while (line != ''):
				self._logs.append(line)
			file_logs.close()
		except:
			raise DirectoryNotFound(self._directory)

		file_logs.close()
		return True
	
	def log(self, server, database, message):
		'''
			(Directory, Server, Database, string) -> (boolean)
			logs the paramaters along with a time-stamp (datetime.now() function) with 
			the _directory text file as well as the _logs class file
			
			*WILL ALSO INCREMEMENT THE COMMIT COUNTER FOR THE LOG SESSIONS*
			
			@paramaters the _directory string is a real file with the correct path
			@exception the function will throw a DirectoryNotFoundError
		'''
		current_time = datetime.now()
		line = f'<{server}<>{database}>?{current_time}?!{message}!'
		try:
			file_logs = open(self._directory, 'a')
			file_logs.write(line)
			self._logs.append(line)
			self._commits += 1
			file_logs.close()
		except:
			raise DirectoryNotFound(self._directory)
		
		return True
	
	def index(self, n):
		'''
			(Directory, int) -> (string)
			indexes the class variable _logs and returns the associative string
			
			@paramaters n >= 0 and n <= length of the log file
			@returns the logged string at index n of the entire logs for a certain directory
			@exception if the index n is invalid, will return IndexError
		'''
		try:
			return self._logs[n]
		except:
			raise IndexError
	
	def lookup(self, column, element, start=0):
		'''
			(Directory, Column, string, int*optional*) -> (string)
			looks up the column element in the logs: RUNTIME (N)
			
			@paramaters Column must be a valid enum:
					1) SERVER
					2) DATABASE
					3) TIMESTAMP
					4) MESSAGE
			@returns the entire log of the first index
					 where the column and line match
			@exception if none is found, will throw LogNotFound
					   exception to the calling file
		'''
		for i in range( start, len(self._logs) ):
			element_parsed = self.parse(self._logs[i], column)
			if (self.parse(self._logs[i], column) == element ):
				return self._logs[i]
		raise LogNotFound(self._directory)
	
	def __retr__(self):
		'''
			(Directory) -> (list of strings)
			Returns a list of strings representing the entire log file from txt
		'''
		return self._logs
	
	def __str__(self):
		'''
			(Directory) -> (string)
			Returns the a visualy apealing string representation of the Logger
		'''
		return f'Logger: file=\'{self._directory}\', pushes=\'{self._commits}\''
			
		


from pyquicksql.utils.sql_errors import *
from pyquicksql.utils.sql_enums import *

import pymysql as sql

def retreive(data_required, data_specific, unknown_column, known_colum, known_element):
	'''
		(list of strings, list of strings, string, string, string) -> (string)
		@conditions valid data_required and data_specific list of strings must be passed as
					arguments to the fucntion in acordance to the outlines bellow. Furthermore,
					a valid know/unknown column and known_data must be entered as arguments to
					the function.
		@returns the data associated with the unknown_column that is found within the row of the
				 known colum and known data.
		@exception if invalid input was encountered, the function will return False with the assumption
				   there is a failure to acknowledge or follow the outlines by the docstrings above
				
			*** SHARED AMONGST VARIOUS DATABASE CALLS : ONE SERVER (MOST LIKELY) ***
			data_required = [ =
				0 : IP ADDRESS OF DATABASE
				1 : PORT OF DATABASE
				2 : USERNAME FOR DB ACESS
				3 : PASSWORD FOR DB ACESS
			]
		
			*** NOT SPECIFIC TO VARIOUS CALLS : HIGH LIKELYHOOD FOR VARIATIONS AMONGST CALLS ***
			data_specific = [
				0 : AUTOCOMMIT (BOOL)
				1 : DB NAME
				2 : DB TABLE NAME
			]
	'''
	connection = sql.connect(
		host = data_required.getIP(),
		port = data_required.getPort(),
		user = data_required.getAccount()[0],
		passwd = data_required.getAccount()[1],
		autocommit = (data_specific.getAutocommit() == 'True'),
		db = data_specific.getName(),
	)
	try:
		with connection.cursor() as cursor:
			#from the table user_lookup, find the data entry where the colum is
			#'ipaddr' representing ip addresses on the database, and retreive the
			#associated user-id found in the row of the given ip-address
			s = f"SELECT {unknown_column} FROM {data_specific.getTable()} WHERE {known_colum} = \'{known_element}\';"
			cursor.execute(s)
			data = cursor.fetchone() #fetch the ipaddress associated with the userid **DO NOT REMOVE**
			#while no changes are being made, make sure that the connection to
			#the sql server is closed and not left hanging (unforseen errors)
			connection.close()
	except:
		raise ElementNotFound()
		return '' #in the case of a total crash, still return an empty string
	finally:
		#we are subscripting a list of two pieces of data, though in the case of an error, there may be
		#a return type of None, incase the except is bypassed return an empty string representing that
		#no id could be associated, otherwise subscript the id-address which is the only useful data
		try:
			return data[0]
		except:
			return ''


from pyquicksql.utils.sql_errors import *
from pyquicksql.utils.sql_enums import *
import sqlite3

def commit(data_required, data_specific, column, element):
	'''
		(list of strings, list of strings, string, string) -> (boolean)
		@conditions valid data_required and data_specific list of strings must be passed as
					arguments to the fucntion in acordance to the outlines bellow. Furthermore,
					a valid column and element must be entered as arguments to the function.
		@returns True if the element was sucesfully removed form the specified column of a 
				 pre-existing row to the database without interuption.
		@exception if invalid input was encountered, the function will return False with the assumption
				   there is a failure to acknowledge or follow the outlines by the docstrings above
				
			*** SHARED AMONGST VARIOUS DATABASE CALLS : ONE SERVER (MOST LIKELY) ***
			data_required = [ =
				0 : IP ADDRESS OF DATABASE
				1 : PORT OF DATABASE
				2 : USERNAME FOR DB ACESS
				3 : PASSWORD FOR DB ACESS
			]
		
			*** NOT SPECIFIC TO VARIOUS CALLS : HIGH LIKELYHOOD FOR VARIATIONS AMONGST CALLS ***
			data_specific = [
				0 : AUTOCOMMIT (BOOL)
				1 : DB NAME
				2 : DB TABLE NAME
			]
	'''
	#we want to check to make sure the id is stored in the first place
	#otherwise it is pointless and inefficient to run through the sql
	#procedures for removing the id: aka. fail-safe for future bad programming
	#or steps not-taken into account when this is legacy code
	if (query_lookup.retreive(data_required, data_specific, column, column, element) == ''):
		return False
	
	connection = sql.connect(
		host = data_required.getIP(),
		port = data_required.getPort(),
		user = data_required.getAccount()[0],
		passwd = data_required.getAccount()[1],
		autocommit = (data_specific.getAutocommit() == 'True'),
		db = data_specific.getName(),
	)
	try:
		with connection.cursor() as cursor:
			#command is as follows: where we find that the ip-address which
			#the command is received from, delete that row and it's data from
			#the sql database: we want to offer the ability to remove all cached
			#data to keep anoniminity and user-privacy secured
			s = f'DELETE FROM {data_specific.getTable()} WHERE {column} = \'{element}\';'
			cursor.execute(s)
			#while no changes are being made, make sure that the connection to
			#the sql server is closed and not left hanging (unforseen errors)
			cursor.close()
	except:
		raise ElementNotFound()
		return False
	finally:
		#in the event the code is run safley and without bugs or errors; it is safe to
		#assume and return to the main logical flow that the appropriate requested data
		#has been removed from the coresponding database table
		return True


class Database:
	def __init__(self, _name, _table, _autocommit):
		'''
			(Name, string, int) -> None
			initializes a new database class to be inputed into a connect class
			
			@paramaters a Name enum (MYSQL, SQLITE) must be at least specified
			@exception the default type will be set to MYSQL as it is the more popular choice
			
			** The values cannot be changed, a new database should be create each time **
		'''
		self._name = _name
		self._table = _table
		self._autocommit = _autocommit
	def getName(self):
		'''
			(Database) -> (Name)
			the getter function for the database name string
		'''
		return self._name
	def getTable(self):
		'''
			(Database) -> (string)
			the getter function for the table name string
		'''
		return self._table
	def getAutocommit(self):
		'''
			(Database) -> (boolean)
			the getter function for the boolean value of the autocommit for the table
		'''
		return self._autocommit
	def __eq__(self, other):
		'''
			(Database) -> (boolean)
			compares the current class to other on a basis of the database and table string names
		'''
		if ( self._name != other._name ):
			return False
		if ( self._table != other._table ):
			return False
	def __repr__(self):
		'''
			(Database) -> (string)
			Returns a string representation of the Database class structure
		'''
		return f'Database({self._name}, {self._table}, {self._autocommit})'
	def __str__(self):
		'''
			(Database) -> (string)
			Returns the a visualy apealing string representation of the Database.
		'''
		return f'Database: {self._name} -> Table: {self._table} ({self._autocommit})'


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
	readme_description = fh.read()

desc = 'A light-weight (out-of-the box) tool for pushing SQL (MySQL and SQLite) queries, a markup-language for structured txt files and running data loggers in python.'

setup(
	name = 'pyquicksql',
	packages = find_packages(),
	version = '1.1.1',
	license = 'MIT',
	description = desc,
	author = 'Gabriel Cordovado',
	author_email = 'gabriel.cordovado@icloud.com',
	long_description = readme_description,
	long_description_content_type = 'text/markdown',
	url ='https://github.com/GabeCordo/python-quick-sql',
	download_url = 'https://github.com/GabeCordo/python-quick-sql/archive/v_2.1.tar.gz',
	keywords = ['MYSQL', 'SQLITE', 'LOGGING'],
	install_requires = [
		'cffi',
		'pymysql',
		'cryptography'
	],
	classifiers = [
		'License :: OSI Approved :: MIT License',
		'Operating System :: OS Independent'
	],
	python_requries = '>=3.4',
)